from __future__ import annotations
from pathlib import Path
import typing as T
import logging
import re
import os
from operator import itemgetter
import py_compile
import subprocess

from . import handlers
from . import dotignore

class bcolors:
    HEADER    = '\033[95m'
    OKBLUE    = '\033[94m'
    OKGREEN   = '\033[92m'
    WARNING   = '\033[93m'
    FAIL      = '\033[91m'
    BOLD      = '\033[1m'
    UNDERLINE = '\033[4m'
    ENDC      = '\033[0m'

# http://www.useragentstring.com
USER_AGENT = "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:64.0) Gecko/20100101 Firefox/64.0"
SUMMARY = {
    'total': 0,
    'passed': 0,
    'checked': 0,
    'checked_syntax': 0,
    'checked_runtime': 0,
    'errors': 0,
    'errors_syntax': 0,
    'errors_runtime': 0,
    'skipped': 0,
    'problems': {},
    'failure': False
    }

STATUS_LABELS = {
    'passed':     '✓',
    'permission': '✖',
    'syntax':     '✖',
    'skipped':    '/',
    'error':      '⚠'
}

def print_summary( root_path ):
    SUMMARY["passed_syntax"] = SUMMARY["checked_syntax"] - SUMMARY["errors_syntax"]
    print(f'SUMMARY')
    print(f'  Total files: {SUMMARY["total"]}')
    print(f'SYNTAX CHECKS')
    print(f'    Files: {SUMMARY["checked_syntax"]}')
    print(f'   Passed: {SUMMARY["passed_syntax"]}')
    print(f' Failures: {SUMMARY["errors_syntax"]}')
    print(f'RUNTIME TESTS')
    print(f'    Files: {SUMMARY["checked_runtime"]}')
    print(f'   Passed: {SUMMARY["passed"]}')
    print(f' Failures: {SUMMARY["errors_runtime"]}')

    for file_path in SUMMARY['problems']:
        problem = SUMMARY['problems'][file_path]
        rel_file_path = relative_path( root_path, file_path )
        p_msg = problem['msg']
        p_type = problem['type']
        print(f'\n{bcolors.OKBLUE}{rel_file_path}{bcolors.ENDC}')
        print(f'[{STATUS_LABELS[p_type]}] {p_msg}')

def relative_path( root_path, file_path ) -> str:
    cur_dir_name = str( root_path ).rpartition('/')[2]
    rfp = './' + cur_dir_name + str(file_path).replace( str(root_path), '' )
    return rfp
        
#def ignore_file( fn ) -> bool:
#    # TODO - switch to an .ignore-file file mechanism
#    e = r'^(__init__.py|.*\~|\..*)$'
#    regex = re.compile(e)
#    if regex.search( fn.name ):
#        logging.debug(f'Ignore {fn.name}? Yes.')
#        return True
#    logging.debug(f'Ignore {fn.name}? No.')
#    return False

def append_problem(fn, msg, t):
    #logging.info( fn )
    SUMMARY['problems'][fn["fn"]] = { 'msg': msg, 'type': t }    

def process_code(
        path: Path,
        recurse: bool = False,
        exclude: [str] = None,
        syntax_only: bool = False,
        languages: [str] = None
) -> bool:
    bad = False
    code_files = find_code_samples( path, recurse=recurse, exclude=exclude )
    root_path = None

    logging.debug(f'Processing languages: {languages}')
    for f in code_files:
        full_path = f["fn"].name
        root_path = f["path"]
        SUMMARY['total'] += 1
        logging.debug(f'{SUMMARY["total"]}. Processing {full_path}')
        # This should not be necessary as the dotignore file should be used instead
        #ignore = ignore_file( f["fn"] )
        #logging.debug(f'  {SUMMARY["total"]}. Exclude? {ignore}')
        #if ignore:
        #    continue

        try:
            handler = handlers.find_handler( f )
            logging.debug(f'  {SUMMARY["total"]}. {full_path} is type {handler.language}')
            skip = (languages != None and str(handler.language) not in languages)
            logging.debug(f'  {SUMMARY["total"]}. Skip this file? {skip} (is {handler.language} in {languages})')
            if skip:
                logging.debug(f'  {SUMMARY["total"]}. Skipping language for {handler.language}')
                continue
            else:
                logging.info(f'  {SUMMARY["total"]}. Checking syntax for {full_path}')  
                SUMMARY['checked_syntax'] += 1
                logging.info(f'Checking {handler.language} syntax: {full_path}')
                handler.check_syntax()
                if not syntax_only: 
                    logging.debug(f'  {SUMMARY["total"]}. Executing {full_path}')
                    SUMMARY['checked_runtime'] += 1
                    result = handler.check_runtime()
                    if result.returncode > 0:
                        #logging.info( f'handler.check_runtime() FAILED for {full_path}' )
                        #logging.info( f'   stderr: {result.stderr}' )
                        #logging.info( f'   stdout: {result.stdout}' )
                        #if re.search( r'429 Client Error', result.stderr ):
                        #    logging.info( f'Rate limit error possibly detected.' )
                        append_problem( f, result.stderr, "error" )
                        SUMMARY['errors'] += 1
                        SUMMARY['errors_runtime'] += 1
                    else:
                        #logging.info( f'handler.check_runtime() succeeded. stdout: {result.stdout}' )
                        SUMMARY['passed'] += 1
        except handlers.NoCodeHandler as e:
            logging.debug(f'  {SUMMARY["total"]}. No handler found for: {f["fn"].name}')
            SUMMARY['skipped'] += 1
        except handlers.SyntaxError as e:
            logging.debug(f'  {SUMMARY["total"]}. There is a syntax problem with the file.')
            append_problem( f, f'Syntax error: {e}', 'syntax' )
            SUMMARY['errors_syntax'] += 1
        except handlers.PermissionsError as e:
            logging.debug(f'  {SUMMARY["total"]}. The file is not executable.')
            append_problem( f, "Not executable.", "permission" )
            SUMMARY['errors'] += 1
            SUMMARY['errors_runtime'] += 1
        except handlers.TimedOutError as e:
            logging.debug(f'  {SUMMARY["total"]}. Process took too long to run.')
            append_problem( f, 'Timed out.', 'error' )
            SUMMARY['errors'] += 1
            SUMMARY['errors_runtime'] += 1
        except handlers.RuntimeError as e:
            logging.debug(f'  {SUMMARY["total"]}. The script ({full_path}) exited with an error status code')
            append_problem( f, f'Error executing script: {e}', 'error' )
            SUMMARY['errors'] += 1
            SUMMARY['errors_runtime'] += 1
            
    if SUMMARY['errors'] > 0: bad = True
    print_summary( root_path )
    return bad

def exclude_file( link, exclude ) -> bool:
    for e in exclude:
        regex = re.compile(e)
        if regex.search( link ):
            return True
    return False

def find_code_samples( path: Path, recurse: bool, exclude: [] = None
                  ) -> tuple[ T.Iterable, T.Iterable ]:
    path = Path(path).resolve().expanduser()  # must have .resolve()
    # these look for local and relative links only
    # markdown regex to extract links from [Link label](link url)
    code_samples = []
    di = dotignore.dotignore('.codecheck-ignore')
    for fn in di.get_files(path, recurse):
        #full_path = os.path.join( path, fn.name )
        if os.path.isfile( fn ):
            #logging.debug(f'{fn} is a file')
            code_samples.append( {
                'fn': fn,
                'path': path } )
    return code_samples
                



import logging
import py_compile
import subprocess
import os

def rchop(s, suffix):
    if suffix and s.endswith(suffix):
        return s[:-len(suffix)]
    return s

def is_handler_enabled( language ) -> bool:
    if language == 'python':
        return PythonCodeHandler.is_enabled()
    elif language == 'java':
        return JavaCodeHandler.is_enabled()
    elif language == 'php':
        return PHPCodeHandler.is_enabled()
    elif language == 'ruby':
        return RubyCodeHandler.is_enabled()
    elif language == 'javascript':
        return JavaScriptCodeHandler.is_enabled()
    elif language == 'c#':
        return CSharpCodeHandler.is_enabled()
    else:
        raise UnknownLanguage(f'Unknown language {language}: cannot process.')

def find_handler( f ):# -> CodeHandler:
    logging.debug(f'Finding handler for {f}')
    if PythonCodeHandler.can_handle( f ):
        #logging.info('Python file found')
        return PythonCodeHandler( f )
    elif PHPCodeHandler.can_handle( f ):
        #logging.info('PHP file found')
        return PHPCodeHandler( f )
    elif JavaCodeHandler.can_handle( f ):
        #logging.info('Java file found')
        return JavaCodeHandler( f )
    elif JavaScriptCodeHandler.can_handle( f ):
        #logging.info('JavaScript file found')
        return JavaScriptCodeHandler( f )
    elif RubyCodeHandler.can_handle( f ):
        #logging.info('Ruby file found')
        return RubyCodeHandler( f )
    elif CSharpCodeHandler.can_handle( f ):
        #logging.info('Ruby file found')
        return CSharpCodeHandler( f )
    else:
        #logging.info(f"Could not find handler for {f['fn']}")
        raise NoCodeHandler(f'Could not find handler for {f}')

class CodeHandlerException(Exception):
    """Base class for other exceptions"""
    pass
class NoCodeHandler(CodeHandlerException):
    pass
class UnknownLanguage(CodeHandlerException):
    pass
class SyntaxError(CodeHandlerException):
    pass
class RuntimeError(CodeHandlerException):
    pass
class PermissionsError(CodeHandlerException):
    pass
class TimedOutError(CodeHandlerException):
    pass

class CodeHandler:
    language = None
    code_file = None
    def __init__(self, l, f):
        self.data = []
        self.language = l
        self.code_file = f
    def can_handle( f ) -> bool:
        #logging.info(f'Don\'t know how to detect files for {self.language}')
        pass
    def is_enabled() -> bool:
        #logging.info(f'Don\'t know how to detect files for {self.language}')
        pass
    def check_syntax( self ):
        #logging.info(f'Don\'t know how to check syntax for {self.language}')
        pass
    def check_runtime( self ):
        #logging.info(f'Don\'t know how to check run time for {self.language}')
        pass

class PythonCodeHandler( CodeHandler ):
    def __init__(self, f):
        super().__init__( 'python', f )
        self.data = []
    def can_handle( f ):
        if f["fn"].name.endswith('.py'):
            return True
        return False
    def is_enabled():
        return True
    def check_syntax(self):
        # TODO - capture STDOUT and save full stacktrace to SUMMARY
        super().check_syntax()
        full_path = self.code_file['fn']
        try:
            result = py_compile.compile( full_path, doraise=True )
        except (py_compile.PyCompileError) as e:
            raise SyntaxError(e)
        else:
            return result
    def check_runtime(self):
        # TODO - capture STDOUT and do not output to terminal
        super().check_runtime()
        full_path = self.code_file['fn']
        #logging.info(f'Processing Python file: {full_path}')
        try:
            result = subprocess.run(full_path,stdout=subprocess.PIPE,stderr=subprocess.PIPE,
                                    universal_newlines=True,timeout=10,
                                    check=False)
        except (py_compile.PyCompileError) as e:
            raise SyntaxError(e)
        except OSError as e:
            raise RuntimeError(e)
        except PermissionError as e:
            raise PermissionsError(e)
        except subprocess.TimeoutExpired as e:
            raise TimedOutError(e, result)
        except subprocess.CalledProcessError as e:
            # This is a controlled failure, meaning the script exited with a status > 0
            raise RuntimeError(e)
        else:
            #logging.info( f'Exiting check_runtime() successfully for {full_path}')
            return result

class PHPCodeHandler( CodeHandler ):
    def __init__(self, f):
        super().__init__( 'php', f )
        self.data = []
    def can_handle( f ):
        if f["fn"].name.endswith('.php'):
            return True
        return False
    def is_enabled():
        try:
            result = subprocess.run(['php','-v'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            return True
        except:
            return False
    def check_syntax(self):
        # TODO - capture STDOUT and save full stacktrace to SUMMARY
        super().check_syntax()
        full_path = self.code_file['fn']
        result = subprocess.call(['php','-l',full_path])
        if result != 0:
            raise SyntaxError(f'Syntax error in: {full_path}', result)
        return result
    def check_runtime(self):
        # TODO - capture STDOUT and do not output to terminal
        super().check_runtime()
        full_path = self.code_file['fn']
        #logging.info(f'Processing PHP file: {full_path}')
        result = subprocess.call(['php',full_path])
        if result != 0:
            raise RuntimeError(r'Error running command: php {full_path}')
        return result

class JavaScriptCodeHandler( CodeHandler ):
    def __init__(self, f):
        super().__init__( 'javascript', f )
        self.data = []
    def can_handle( f ):
        if f["fn"].name.endswith('.js') or f["fn"].name.endswith('.json'):
            return True
        return False
    def is_enabled():
        try:
            result = subprocess.run(['node','-v'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            return True
        except:
            return False
    def check_syntax(self):
        # TODO - capture STDOUT and save full stacktrace to SUMMARY
        super().check_syntax()
        full_path = self.code_file['fn']
        result = subprocess.run(['node','--check',full_path], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return result
    def check_runtime(self):
        # TODO - capture STDOUT and do not output to terminal
        super().check_runtime()
        full_path = self.code_file['fn']
        #logging.info(f'Processing JavaScript file: {full_path}')
        result = subprocess.run(['node',full_path], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return result

class RubyCodeHandler( CodeHandler ):
    def __init__(self, f):
        super().__init__( 'ruby', f )
        self.data = []
    def can_handle( f ):
        if f["fn"].name.endswith('.rb'):
            return True
        return False
    def is_enabled():
        try:
            result = subprocess.run(['ruby','-v'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            return True
        except:
            return False
    def check_syntax(self):
        # TODO - capture STDOUT and save full stacktrace to SUMMARY
        super().check_syntax()
        full_path = self.code_file['fn']
        result = subprocess.call(['ruby','-c',full_path])
        if result != 0:
            raise SyntaxError(f'Syntax error in: {full_path}', result)
        return result
    def check_runtime(self):
        # TODO - capture STDOUT and do not output to terminal
        super().check_runtime()
        full_path = self.code_file['fn']
        #logging.info(f'Processing JavaScript file: {full_path}')
        result = subprocess.call(['ruby',full_path])
        if result != 0:
            raise RuntimeError(r'Error running command: ruby {full_path}')
        return result

class JavaCodeHandler( CodeHandler ):
    def __init__(self, f):
        super().__init__( 'java', f )
        self.data = []
    def can_handle( f ):
        if f["fn"].name.endswith('.java'):
            return True
        return False
    def is_enabled():
        try:
            result1 = subprocess.run(['java','--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            result2 = subprocess.run(['javac','--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            return True
        except:
            return False
    def check_syntax(self):
        # TODO - capture STDOUT and save full stacktrace to SUMMARY
        super().check_syntax()
        full_path = self.code_file['fn']
        result = subprocess.call(['javac',full_path])
        if result != 0:
            raise SyntaxError(f'Syntax error in: {full_path}', result)
        return result
    def check_runtime(self):
        # TODO - capture STDOUT and do not output to terminal
        super().check_runtime()
        full_path = self.code_file['fn']
        #logging.info(f'Processing Java file: {full_path}')
        #return result

class CSharpCodeHandler( CodeHandler ):
    def __init__(self, f):
        super().__init__( 'csharp', f )
        self.data = []
    def can_handle( f ):
        if f["fn"].name == 'Program.cs':
            return True
        return False
    def is_enabled():
        try:
            result1 = subprocess.run(['dotnet','--info'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            return True
        except:
            return False
    def check_syntax(self):
        # TODO - capture STDOUT and save full stacktrace to SUMMARY
        super().check_syntax()
        full_path = self.code_file['fn']
        prj_dir = rchop(str(full_path), 'Program.cs')
        result = subprocess.call(['dotnet','build',prj_dir])
        if result != 0:
            raise SyntaxError(f'Syntax error in: {full_path}', result)
        return result
    def check_runtime(self):
        # TODO - capture STDOUT and do not output to terminal
        super().check_runtime()
        full_path = self.code_file['fn']
        prj_dir = rchop(str(full_path), 'Program.cs')
        result = subprocess.call(['dotnet','run',prj_dir])
        if result != 0:
            raise RuntimeError(f'Error running project: {prj_dir}', result)
        return result




"""
Check code samples in a directory tree.

% mkdocs-codecheck ~/mySite/code-samples
"""

import argparse
import logging
import time
from dotenv import load_dotenv
from pathlib import Path

from .base import process_code
from . import handlers

def main():
    p = argparse.ArgumentParser(description="Check code files within a directory or tree.")
    p.add_argument(
        "path",
        help="Path to code samples directory")
    p.add_argument(
        "-v", "--verbose",
        action="store_true")
    p.add_argument(
        "--exclude",
        help="a pattern for a file or path to exclude",
        action="append")
    p.add_argument(
        "-r", "--recurse",
        help="recurse directories under path",
        action="store_true")
    p.add_argument(
        "--syntax-only",
        help="Check syntax of code only. Do not execute the script.",
        action="store_true")
    p.add_argument(
        "--languages",
        help="The languages to process. A comma delimitted list of any of java, python, php, ruby, javascript, and c#.")
    p.add_argument(
        "--dotenv",
        help="The path to a .env file that contains environment variables to pull into the current execution context")
    P = p.parse_args()

    if P.verbose:
        logging.basicConfig(level=logging.INFO)

    if P.dotenv:
        dotenv_path = Path( P.dotenv )
        load_dotenv(dotenv_path=dotenv_path)
        
    tic = time.monotonic()
    langs = []
    if P.languages != None:
        for l in P.languages.split(','):
            l = l.strip()
            try:
                if not handlers.is_handler_enabled( l ):
                    print(f'Error: environment not setup properly for {l}, bailing.')
                    raise SystemExit(22)
            except handlers.UnknownLanguage as e:
                print(f'Error: unknown language "{l}"')
                raise SystemExit(22)
            langs.append(l)
    bad = process_code(
        P.path,
        recurse=P.recurse,
        exclude=P.exclude,
        syntax_only=P.syntax_only,
        languages=langs
    )

    print(f"\n{time.monotonic() - tic:0.3} seconds to check code samples")

    if bad:
        # using 22 following cURL
        # https://everything.curl.dev/usingcurl/returns
        print("Errors were discovered in your code samples. Exiting with an error.")
        raise SystemExit(22)


if __name__ == "__main__":
    main()



from pathlib import Path
import typing as T
import logging
import re

logging.basicConfig(level=logging.INFO)

class dotignore:
    
    def __init__(self, filename):
        self.rules = []
        self.dotignore_filename = filename

    def read_dotignore( self ):
        """
        This will be throwing something most likely in an error. How do we handle that?
        """
        logging.debug(f'Opening {self.dotignore_filename}')
        with open(self.dotignore_filename) as file:
            lines = file.readlines()
            for line in lines:
                line = line.rstrip()
                logging.debug(f'line: {line}')
                self.rules.append( line )
        return self.rules

    def ignore_file( self, filename ):
        m = re.match( r'^(.+)\/([^\/]+)$', str(filename) )
        path = m.group(1)
        fn = m.group(2)
        logging.debug(f'path={path}, file={fn}')
        for rule in self.rules:
            rule = str(rule)
            logging.debug(f'Checking {fn} against {rule}')
            if re.search( rf"{rule}", str(filename), re.IGNORECASE ):
                logging.debug(f'Ignoring {fn}')
                return True
        return False
        
    def get_files(self, path: Path, recurse: bool = False) -> T.Iterable[Path]:
        """
        yield files in path with suffix ext. Optionally, recurse directories.
        """
        self.read_dotignore()
        path = Path(path).expanduser().resolve()
        if path.is_dir():
            for p in path.iterdir():
                if p.is_file():
                    if self.ignore_file( p ):
                        continue
                    else:
                        yield p
                elif p.is_dir():
                    if recurse:
                        yield from self.get_files(p, recurse)
        elif path.is_file():
            logging.debug(f'Adding file to be checked: {path}')
            yield path
        else:
            raise FileNotFoundError(path)

